from __future__ import annotations

import argparse
import enum
from datetime import datetime, timezone
from pathlib import Path
from typing import Any

import deltalake
import duckdb
import pandas as pd
import pyarrow as pa
import pydantic
import rich.panel
import rich.table
import rich.text
import rich.tree
import sqlglot
import sqlglot.dialects
import sqlglot.dialects.duckdb
import sqlglot.generator
import yaml

from laketower.__about__ import __version__


class TableFormats(str, enum.Enum):
    delta = "delta"


class ConfigTable(pydantic.BaseModel):
    name: str
    uri: str
    table_format: TableFormats = pydantic.Field(alias="format")

    @pydantic.model_validator(mode="after")
    def check_table(self) -> "ConfigTable":
        def check_delta_table(table_uri: str) -> None:
            if not deltalake.DeltaTable.is_deltatable(table_uri):
                raise ValueError(f"{table_uri} is not a valid Delta table")

        format_check = {TableFormats.delta: check_delta_table}
        format_check[self.table_format](self.uri)

        return self


class ConfigQuery(pydantic.BaseModel):
    name: str
    sql: str


class ConfigDashboard(pydantic.BaseModel):
    name: str


class Config(pydantic.BaseModel):
    tables: list[ConfigTable] = []


def load_yaml_config(config_path: Path) -> Config:
    config_dict = yaml.safe_load(config_path.read_text())
    return Config.model_validate(config_dict)


class TableMetadata(pydantic.BaseModel):
    table_format: TableFormats
    name: str
    description: str
    uri: str
    id: str
    version: int
    created_at: datetime
    partitions: list[str]
    configuration: dict[str, str]


class TableRevision(pydantic.BaseModel):
    version: int
    timestamp: datetime
    client_version: str
    operation: str
    operation_parameters: dict[str, Any]
    operation_metrics: dict[str, Any]


class TableHistory(pydantic.BaseModel):
    revisions: list[TableRevision]


def load_table_metadata(table_config: ConfigTable) -> TableMetadata:
    def load_delta_table_metadata(table_config: ConfigTable) -> TableMetadata:
        delta_table = deltalake.DeltaTable(table_config.uri)
        metadata = delta_table.metadata()
        return TableMetadata(
            table_format=table_config.table_format,
            name=metadata.name,
            description=metadata.description,
            uri=delta_table.table_uri,
            id=str(metadata.id),
            version=delta_table.version(),
            created_at=datetime.fromtimestamp(
                metadata.created_time / 1000, tz=timezone.utc
            ),
            partitions=metadata.partition_columns,
            configuration=metadata.configuration,
        )

    format_handler = {TableFormats.delta: load_delta_table_metadata}
    return format_handler[table_config.table_format](table_config)


def load_table_schema(table_config: ConfigTable) -> pa.Schema:
    def load_delta_table_schema(table_config: ConfigTable) -> pa.Schema:
        delta_table = deltalake.DeltaTable(table_config.uri)
        return delta_table.schema().to_pyarrow()

    format_handler = {TableFormats.delta: load_delta_table_schema}
    return format_handler[table_config.table_format](table_config)


def load_table_history(table_config: ConfigTable) -> TableHistory:
    def load_delta_table_history(table_config: ConfigTable) -> TableHistory:
        delta_table = deltalake.DeltaTable(table_config.uri)
        delta_history = delta_table.history()
        revisions = [
            TableRevision(
                version=event["version"],
                timestamp=datetime.fromtimestamp(
                    event["timestamp"] / 1000, tz=timezone.utc
                ),
                client_version=event["clientVersion"],
                operation=event["operation"],
                operation_parameters=event["operationParameters"],
                operation_metrics=event.get("operationMetrics") or {},
            )
            for event in delta_history
        ]
        return TableHistory(revisions=revisions)

    format_handler = {TableFormats.delta: load_delta_table_history}
    return format_handler[table_config.table_format](table_config)


def load_table_dataset(table_config: ConfigTable) -> pa.dataset.Dataset:
    def load_delta_table_metadata(table_config: ConfigTable) -> pa.dataset.Dataset:
        delta_table = deltalake.DeltaTable(table_config.uri)
        return delta_table.to_pyarrow_dataset()

    format_handler = {TableFormats.delta: load_delta_table_metadata}
    return format_handler[table_config.table_format](table_config)


def execute_query_table(table_config: ConfigTable, sql_query: str) -> pd.DataFrame:
    table_dataset = load_table_dataset(table_config)
    table_name = table_config.name
    view_name = f"{table_name}_view"
    conn = duckdb.connect()
    conn.register(view_name, table_dataset)
    conn.execute(f"create table {table_name} as select * from {view_name}")  # nosec B608
    return conn.execute(sql_query).df()


def execute_query(tables_config: list[ConfigTable], sql_query: str) -> pd.DataFrame:
    try:
        conn = duckdb.connect()
        for table_config in tables_config:
            table_dataset = load_table_dataset(table_config)
            table_name = table_config.name
            view_name = f"{table_name}_view"
            conn.register(view_name, table_dataset)
            conn.execute(f"create table {table_name} as select * from {view_name}")  # nosec B608
        return conn.execute(sql_query).df()
    except duckdb.Error as e:
        raise ValueError(str(e)) from e


def validate_config(config_path: Path) -> None:
    console = rich.get_console()
    try:
        config = load_yaml_config(config_path)
        console.print(rich.panel.Panel.fit("[green]Configuration is valid"))
        console.print(config)
    except Exception as e:
        console.print(rich.panel.Panel.fit("[red]Configuration is invalid"))
        console.print(e)


def list_tables(config_path: Path) -> None:
    config = load_yaml_config(config_path)
    tree = rich.tree.Tree("tables")
    for table in config.tables:
        table_tree = tree.add(table.name)
        table_tree.add(f"format: {table.table_format.value}")
        table_tree.add(f"uri: {table.uri}")
    console = rich.get_console()
    console.print(tree)


def table_metadata(config_path: Path, table_name: str) -> None:
    config = load_yaml_config(config_path)
    table_config = next(filter(lambda x: x.name == table_name, config.tables))
    metadata = load_table_metadata(table_config)

    tree = rich.tree.Tree(table_name)
    tree.add(f"name: {metadata.name}")
    tree.add(f"description: {metadata.description}")
    tree.add(f"format: {metadata.table_format.value}")
    tree.add(f"uri: {metadata.uri}")
    tree.add(f"id: {metadata.id}")
    tree.add(f"version: {metadata.version}")
    tree.add(f"created at: {metadata.created_at}")
    tree.add(f"partitions: {', '.join(metadata.partitions)}")
    tree.add(f"configuration: {metadata.configuration}")
    console = rich.get_console()
    console.print(tree)


def table_schema(config_path: Path, table_name: str) -> None:
    config = load_yaml_config(config_path)
    table_config = next(filter(lambda x: x.name == table_name, config.tables))
    schema = load_table_schema(table_config)

    tree = rich.tree.Tree(table_name)
    for field in schema:
        nullable = "" if field.nullable else " not null"
        tree.add(f"{field.name}: {field.type}{nullable}")
    console = rich.get_console()
    console.print(tree, markup=False)  # disable markup to allow bracket characters


def table_history(config_path: Path, table_name: str) -> None:
    config = load_yaml_config(config_path)
    table_config = next(filter(lambda x: x.name == table_name, config.tables))
    history = load_table_history(table_config)

    tree = rich.tree.Tree(table_name)
    for rev in history.revisions:
        tree_version = tree.add(f"version: {rev.version}")
        tree_version.add(f"timestamp: {rev.timestamp}")
        tree_version.add(f"client version: {rev.client_version}")
        tree_version.add(f"operation: {rev.operation}")
        tree_op_params = tree_version.add("operation parameters")
        for param_key, param_val in rev.operation_parameters.items():
            tree_op_params.add(f"{param_key}: {param_val}")
        tree_op_metrics = tree_version.add("operation metrics")
        for metric_key, metric_val in rev.operation_metrics.items():
            tree_op_metrics.add(f"{metric_key}: {metric_val}")
    console = rich.get_console()
    console.print(tree, markup=False)


def view_table(
    config_path: Path,
    table_name: str,
    limit: int | None = None,
    cols: list[str] | None = None,
    sort_asc: str | None = None,
    sort_desc: str | None = None,
) -> None:
    config = load_yaml_config(config_path)
    table_config = next(filter(lambda x: x.name == table_name, config.tables))

    query_expr = sqlglot.select(*(cols or ["*"])).from_(table_name).limit(limit or 10)
    if sort_asc:
        query_expr = query_expr.order_by(f"{sort_asc} asc")
    elif sort_desc:
        query_expr = query_expr.order_by(f"{sort_desc} desc")
    sql_query = sqlglot.Generator(dialect=sqlglot.dialects.DuckDB).generate(query_expr)

    results = execute_query_table(table_config, sql_query)
    out = rich.table.Table()
    for column in results.columns:
        out.add_column(column)
    for value_list in results.values.tolist():
        row = [str(x) for x in value_list]
        out.add_row(*row)

    console = rich.get_console()
    console.print(out)


def query_table(config_path: Path, sql_query: str) -> None:
    config = load_yaml_config(config_path)

    out: rich.jupyter.JupyterMixin
    try:
        results = execute_query(config.tables, sql_query)
        out = rich.table.Table()
        for column in results.columns:
            out.add_column(column)
        for value_list in results.values.tolist():
            row = [str(x) for x in value_list]
            out.add_row(*row)
    except ValueError as e:
        out = rich.panel.Panel.fit(f"[red]{e}")

    console = rich.get_console()
    console.print(out)


def cli() -> None:
    parser = argparse.ArgumentParser("laketower")
    parser.add_argument("--version", action="version", version=__version__)
    parser.add_argument(
        "--config",
        "-c",
        default="laketower.yml",
        type=Path,
        help="Path to the Laketower YAML configuration file",
    )
    subparsers = parser.add_subparsers(title="commands", required=True)

    parser_config = subparsers.add_parser(
        "config", help="Work with configuration", add_help=True
    )
    subsparsers_config = parser_config.add_subparsers(required=True)

    parser_config_validate = subsparsers_config.add_parser(
        "validate", help="Validate YAML configuration"
    )
    parser_config_validate.set_defaults(func=lambda x: validate_config(x.config))

    parser_tables = subparsers.add_parser("tables", help="Work with tables")
    subsparsers_tables = parser_tables.add_subparsers(required=True)

    parser_tables_list = subsparsers_tables.add_parser(
        "list", help="List all registered tables"
    )
    parser_tables_list.set_defaults(func=lambda x: list_tables(x.config))

    parser_tables_metadata = subsparsers_tables.add_parser(
        "metadata", help="Display a given table metadata"
    )
    parser_tables_metadata.add_argument("table", help="Name of the table")
    parser_tables_metadata.set_defaults(
        func=lambda x: table_metadata(x.config, x.table)
    )

    parser_tables_schema = subsparsers_tables.add_parser(
        "schema", help="Display a given table schema"
    )
    parser_tables_schema.add_argument("table", help="Name of the table")
    parser_tables_schema.set_defaults(func=lambda x: table_schema(x.config, x.table))

    parser_tables_history = subsparsers_tables.add_parser(
        "history", help="Display the history of a given table schema"
    )
    parser_tables_history.add_argument("table", help="Name of the table")
    parser_tables_history.set_defaults(func=lambda x: table_history(x.config, x.table))

    parser_tables_view = subsparsers_tables.add_parser(
        "view", help="View a given table"
    )
    parser_tables_view.add_argument("table", help="Name of the table")
    parser_tables_view.add_argument(
        "--limit", type=int, help="Maximum number of rows to display"
    )
    parser_tables_view.add_argument("--cols", nargs="*", help="Columns to display")
    parser_tables_view_sort_group = parser_tables_view.add_mutually_exclusive_group()
    parser_tables_view_sort_group.add_argument(
        "--sort-asc", help="Sort by given column in ascending order"
    )
    parser_tables_view_sort_group.add_argument(
        "--sort-desc", help="Sort by given column in descending order"
    )
    parser_tables_view.set_defaults(
        func=lambda x: view_table(
            x.config, x.table, x.limit, x.cols, x.sort_asc, x.sort_desc
        )
    )

    parser_tables_query = subsparsers_tables.add_parser(
        "query", help="Query registered tables"
    )
    parser_tables_query.add_argument("sql", help="SQL query to execute")
    parser_tables_query.set_defaults(func=lambda x: query_table(x.config, x.sql))

    args = parser.parse_args()
    args.func(args)



from invoke import task
from invoke.context import Context

app_path = "laketower"
tests_path = "tests"


@task
def format(ctx: Context) -> None:
    ctx.run("ruff format .", echo=True, pty=True)


@task
def audit(ctx):
    ctx.run("pip-audit", echo=True, pty=True)


@task
def vuln(ctx):
    ctx.run(f"bandit -r {app_path}", echo=True, pty=True)


@task
def lint(ctx: Context) -> None:
    ctx.run("ruff check .", echo=True, pty=True)


@task
def typing(ctx: Context) -> None:
    ctx.run(f"mypy --strict {app_path} {tests_path}", echo=True, pty=True)


@task
def test(ctx):
    ctx.run(
        f"py.test -v --cov={app_path} --cov={tests_path} --cov-branch --cov-report=term-missing {tests_path}",
        echo=True,
        pty=True,
    )


@task(audit, vuln, lint, typing, test)
def qa(ctx):
    pass



import sys
from datetime import datetime, timezone
from pathlib import Path
from typing import Any

import deltalake
import pandas as pd
import pytest
import yaml

from laketower import cli


@pytest.fixture()
def delta_table_data() -> pd.DataFrame:
    return pd.DataFrame(
        [
            {
                "time": "2025-01-01T00:00:00+00:00",
                "city": "Grenoble",
                "temperature": "-2",
            },
            {
                "time": "2025-01-01T01:00:00+00:00",
                "city": "Grenoble",
                "temperature": "-3",
            },
            {
                "time": "2025-01-01T02:00:00+00:00",
                "city": "Grenoble",
                "temperature": "-4",
            },
            {
                "time": "2025-01-01T03:00:00+00:00",
                "city": "Grenoble",
                "temperature": "-1",
            },
            {
                "time": "2025-01-01T04:00:00+00:00",
                "city": "Grenoble",
                "temperature": "0",
            },
        ]
    )


@pytest.fixture()
def delta_table(tmp_path: Path, delta_table_data: pd.DataFrame) -> deltalake.DeltaTable:
    table_path = tmp_path / "delta_table"
    schema = deltalake.Schema(
        [
            deltalake.Field("time", "timestamp_ntz", nullable=False),  # type: ignore[arg-type]
            deltalake.Field("city", "string", nullable=False),  # type: ignore[arg-type]
            deltalake.Field("temperature", "float", nullable=False),  # type: ignore[arg-type]
        ]
    )
    dt = deltalake.DeltaTable.create(
        table_path, schema, name="delta_table", description="Sample Delta Table"
    )
    deltalake.write_deltalake(dt, delta_table_data, mode="append")
    return dt


@pytest.fixture()
def sample_config(delta_table: deltalake.DeltaTable) -> dict[str, Any]:
    return {
        "tables": [
            {"name": "delta_table", "uri": delta_table.table_uri, "format": "delta"}
        ]
    }


@pytest.fixture()
def sample_config_path(tmp_path: Path, sample_config: dict[str, Any]) -> Path:
    config_path = tmp_path / "laketower.yml"
    config_path.write_text(yaml.dump(sample_config))
    return config_path


def test_version(
    monkeypatch: pytest.MonkeyPatch, capsys: pytest.CaptureFixture[str]
) -> None:
    from laketower.__about__ import __version__

    monkeypatch.setattr(
        sys,
        "argv",
        ["laketower", "--version"],
    )

    with pytest.raises(SystemExit):
        cli.cli()

    captured = capsys.readouterr()
    output = captured.out
    assert __version__ in output


def test_config_validate(
    monkeypatch: pytest.MonkeyPatch,
    capsys: pytest.CaptureFixture[str],
    sample_config_path: Path,
) -> None:
    monkeypatch.setattr(
        sys,
        "argv",
        ["laketower", "--config", str(sample_config_path), "config", "validate"],
    )

    cli.cli()

    captured = capsys.readouterr()
    output = captured.out
    assert "Configuration is valid" in output


def test_config_validate_unknown_path(
    monkeypatch: pytest.MonkeyPatch, capsys: pytest.CaptureFixture[str]
) -> None:
    monkeypatch.setattr(
        sys, "argv", ["laketower", "--config", "unknown.yml", "config", "validate"]
    )

    cli.cli()

    captured = capsys.readouterr()
    output = captured.out
    assert "Configuration is invalid" in output


def test_config_validate_invalid_table_format(
    monkeypatch: pytest.MonkeyPatch,
    capsys: pytest.CaptureFixture[str],
    tmp_path: Path,
    sample_config: dict[str, Any],
) -> None:
    sample_config["tables"][0]["format"] = "unknown_format"
    sample_config_path = tmp_path / "laketower.yml"
    sample_config_path.write_text(yaml.dump(sample_config))

    monkeypatch.setattr(
        sys,
        "argv",
        ["laketower", "--config", str(sample_config_path), "config", "validate"],
    )

    cli.cli()

    captured = capsys.readouterr()
    output = captured.out
    assert "Configuration is invalid" in output


def test_config_validate_invalid_delta_table(
    monkeypatch: pytest.MonkeyPatch,
    capsys: pytest.CaptureFixture[str],
    tmp_path: Path,
    sample_config: dict[str, Any],
) -> None:
    sample_config["tables"][0]["uri"] = "unknown_table"
    sample_config_path = tmp_path / "laketower.yml"
    sample_config_path.write_text(yaml.dump(sample_config))

    monkeypatch.setattr(
        sys,
        "argv",
        ["laketower", "--config", str(sample_config_path), "config", "validate"],
    )

    cli.cli()

    captured = capsys.readouterr()
    output = captured.out
    assert "Configuration is invalid" in output


def test_tables_list(
    monkeypatch: pytest.MonkeyPatch,
    capsys: pytest.CaptureFixture[str],
    sample_config: dict[str, Any],
    sample_config_path: Path,
) -> None:
    monkeypatch.setattr(
        sys,
        "argv",
        ["laketower", "--config", str(sample_config_path), "tables", "list"],
    )

    cli.cli()

    captured = capsys.readouterr()
    output = captured.out
    for table in sample_config["tables"]:
        assert table["name"] in output
        assert Path(table["uri"]).name in output
        assert f"format: {table['format']}" in output


def test_tables_metadata(
    monkeypatch: pytest.MonkeyPatch,
    capsys: pytest.CaptureFixture[str],
    sample_config: dict[str, Any],
    sample_config_path: Path,
    delta_table: deltalake.DeltaTable,
) -> None:
    monkeypatch.setattr(
        sys,
        "argv",
        [
            "laketower",
            "--config",
            str(sample_config_path),
            "tables",
            "metadata",
            sample_config["tables"][0]["name"],
        ],
    )

    cli.cli()

    captured = capsys.readouterr()
    output = captured.out
    assert "format: delta" in output
    assert f"name: {delta_table.metadata().name}" in output
    assert f"description: {delta_table.metadata().description}" in output
    assert "uri:" in output
    assert f"id: {delta_table.metadata().id}" in output
    assert f"version: {delta_table.version()}" in output
    assert (
        f"created at: {datetime.fromtimestamp(delta_table.metadata().created_time / 1000, tz=timezone.utc)}"
        in output
    )
    assert (
        f"partitions: {', '.join(delta_table.metadata().partition_columns)}" in output
    )
    assert f"configuration: {delta_table.metadata().configuration}" in output


def test_tables_schema(
    monkeypatch: pytest.MonkeyPatch,
    capsys: pytest.CaptureFixture[str],
    sample_config: dict[str, Any],
    sample_config_path: Path,
    delta_table: deltalake.DeltaTable,
) -> None:
    table_name = sample_config["tables"][0]["name"]

    monkeypatch.setattr(
        sys,
        "argv",
        [
            "laketower",
            "--config",
            str(sample_config_path),
            "tables",
            "schema",
            table_name,
        ],
    )

    cli.cli()

    captured = capsys.readouterr()
    output = captured.out
    assert output.startswith(table_name)
    for field in delta_table.schema().to_pyarrow():
        nullable = "" if field.nullable else " not null"
        assert f"{field.name}: {field.type}{nullable}" in output


def test_tables_history(
    monkeypatch: pytest.MonkeyPatch,
    capsys: pytest.CaptureFixture[str],
    sample_config: dict[str, Any],
    sample_config_path: Path,
    delta_table: deltalake.DeltaTable,
) -> None:
    table_name = sample_config["tables"][0]["name"]

    monkeypatch.setattr(
        sys,
        "argv",
        [
            "laketower",
            "--config",
            str(sample_config_path),
            "tables",
            "history",
            table_name,
        ],
    )

    cli.cli()

    captured = capsys.readouterr()
    output = captured.out
    assert output.startswith(table_name)
    for rev in delta_table.history():
        assert f"version: {rev['version']}" in output
        assert (
            f"timestamp: {datetime.fromtimestamp(rev['timestamp'] / 1000, tz=timezone.utc)}"
            in output
        )
        assert f"client version: {rev['clientVersion']}" in output
        assert f"operation: {rev['operation']}" in output
        assert "operation parameters" in output
        operation_parameters = rev["operationParameters"]
        for param_key in operation_parameters.keys():
            assert f"{param_key}: " in output
        assert "operation metrics" in output
        operation_metrics = rev.get("operationMetrics")
        if operation_metrics:
            for metric_key, metric_val in operation_metrics.items():
                assert f"{metric_key}: {metric_val}" in output


def test_tables_view(
    monkeypatch: pytest.MonkeyPatch,
    capsys: pytest.CaptureFixture[str],
    sample_config: dict[str, Any],
    sample_config_path: Path,
    delta_table: deltalake.DeltaTable,
) -> None:
    default_limit = 10

    monkeypatch.setattr(
        sys,
        "argv",
        [
            "laketower",
            "--config",
            str(sample_config_path),
            "tables",
            "view",
            sample_config["tables"][0]["name"],
        ],
    )

    cli.cli()

    captured = capsys.readouterr()
    output = captured.out
    assert all(field.name in output for field in delta_table.schema().fields)

    df = delta_table.to_pandas()[0:default_limit]
    assert all(str(row[col]) in output for _, row in df.iterrows() for col in row.index)


def test_tables_view_limit(
    monkeypatch: pytest.MonkeyPatch,
    capsys: pytest.CaptureFixture[str],
    sample_config: dict[str, Any],
    sample_config_path: Path,
    delta_table: deltalake.DeltaTable,
) -> None:
    selected_limit = 1

    monkeypatch.setattr(
        sys,
        "argv",
        [
            "laketower",
            "--config",
            str(sample_config_path),
            "tables",
            "view",
            sample_config["tables"][0]["name"],
            "--limit",
            str(selected_limit),
        ],
    )

    cli.cli()

    captured = capsys.readouterr()
    output = captured.out
    assert all(field.name in output for field in delta_table.schema().fields)

    df = delta_table.to_pandas()[0:selected_limit]
    assert all(str(row[col]) in output for _, row in df.iterrows() for col in row.index)


def test_tables_view_cols(
    monkeypatch: pytest.MonkeyPatch,
    capsys: pytest.CaptureFixture[str],
    sample_config: dict[str, Any],
    sample_config_path: Path,
    delta_table: deltalake.DeltaTable,
) -> None:
    num_fields = len(delta_table.schema().fields)
    selected_columns = [
        delta_table.schema().fields[i].name for i in range(num_fields - 1)
    ]
    filtered_columns = [delta_table.schema().fields[num_fields - 1].name]

    monkeypatch.setattr(
        sys,
        "argv",
        [
            "laketower",
            "--config",
            str(sample_config_path),
            "tables",
            "view",
            sample_config["tables"][0]["name"],
            "--cols",
            *selected_columns,
        ],
    )

    cli.cli()

    captured = capsys.readouterr()
    output = captured.out
    assert all(col in output for col in selected_columns)
    assert not all(col in output for col in filtered_columns)

    df = delta_table.to_pandas()
    assert all(str(row) in output for row in df[selected_columns])
    assert not all(str(row) in output for row in df[filtered_columns])


def test_tables_view_sort_asc(
    monkeypatch: pytest.MonkeyPatch,
    capsys: pytest.CaptureFixture[str],
    sample_config: dict[str, Any],
    sample_config_path: Path,
    delta_table: deltalake.DeltaTable,
) -> None:
    sort_column = delta_table.schema().fields[0].name

    monkeypatch.setattr(
        sys,
        "argv",
        [
            "laketower",
            "--config",
            str(sample_config_path),
            "tables",
            "view",
            sample_config["tables"][0]["name"],
            "--sort-asc",
            sort_column,
        ],
    )

    cli.cli()

    captured = capsys.readouterr()
    output = captured.out
    assert all(field.name in output for field in delta_table.schema().fields)

    df = delta_table.to_pandas().sort_values(by=sort_column, ascending=True)
    assert all(str(row[col]) in output for _, row in df.iterrows() for col in row.index)


def test_tables_view_sort_desc(
    monkeypatch: pytest.MonkeyPatch,
    capsys: pytest.CaptureFixture[str],
    sample_config: dict[str, Any],
    sample_config_path: Path,
    delta_table: deltalake.DeltaTable,
) -> None:
    sort_column = delta_table.schema().fields[0].name

    monkeypatch.setattr(
        sys,
        "argv",
        [
            "laketower",
            "--config",
            str(sample_config_path),
            "tables",
            "view",
            sample_config["tables"][0]["name"],
            "--sort-desc",
            sort_column,
        ],
    )

    cli.cli()

    captured = capsys.readouterr()
    output = captured.out
    assert all(field.name in output for field in delta_table.schema().fields)

    df = delta_table.to_pandas().sort_values(by=sort_column, ascending=False)
    assert all(str(row[col]) in output for _, row in df.iterrows() for col in row.index)


def test_tables_query(
    monkeypatch: pytest.MonkeyPatch,
    capsys: pytest.CaptureFixture[str],
    sample_config: dict[str, Any],
    sample_config_path: Path,
    delta_table: deltalake.DeltaTable,
) -> None:
    selected_column = delta_table.schema().fields[0].name
    filtered_columns = [field.name for field in delta_table.schema().fields[1:]]
    selected_limit = 1

    monkeypatch.setattr(
        sys,
        "argv",
        [
            "laketower",
            "--config",
            str(sample_config_path),
            "tables",
            "query",
            f"select {selected_column} from {sample_config['tables'][0]['name']} limit {selected_limit}",
        ],
    )

    cli.cli()

    captured = capsys.readouterr()
    output = captured.out
    assert selected_column in output
    assert not all(col in output for col in filtered_columns)

    df = delta_table.to_pandas()
    assert all(str(row) in output for row in df[selected_column][0:selected_limit])
    assert not all(str(row) in output for row in df[selected_column][selected_limit:])


def test_tables_query_invalid(
    monkeypatch: pytest.MonkeyPatch,
    capsys: pytest.CaptureFixture[str],
    sample_config_path: Path,
    delta_table: deltalake.DeltaTable,
) -> None:
    monkeypatch.setattr(
        sys,
        "argv",
        [
            "laketower",
            "--config",
            str(sample_config_path),
            "tables",
            "query",
            "select * from unknown_table",
        ],
    )

    cli.cli()

    captured = capsys.readouterr()
    output = captured.out
    assert "Error" in output
    assert not all(field.name in output for field in delta_table.schema().fields)

    df = delta_table.to_pandas()
    assert not all(
        str(row[col]) in output for _, row in df.iterrows() for col in row.index
    )
